class FetchCase(object):

    def fetch(self, *args, **kwargs):

        # it's really annoying when Tornado automatically follows redirects.
        # I set this to be false by default, because explicit > implicit.
        # Override by specifying follow_redirects=True

        kwargs.setdefault("follow_redirects", False)
        return super(FetchCase, self).fetch(*args, **kwargs)



from testnado.credentials.helpers import build_fetch_arguments
from testnado.fetch_case import FetchCase


class AuthenticatedFetchCase(FetchCase):

    def authenticated_fetch(
            self, path, method=None, headers=None, body=None,
            auth_username=None, auth_password=None, auth_mode=None, **kwargs):

        headers = headers or {}

        # fetch arguments is the explicit list of overwritable arguments
        # for fetch() calls. For instance, 'method' isn't overwritable,
        # because that would be very bizarre if you could change the method
        # of a test fetch() call just based on the credentials...

        # TODO: change URL to query for query arguments instead...

        fetch_arguments = build_fetch_arguments(
            path=path, headers=headers, body=body, auth_mode=auth_mode,
            auth_username=auth_username, auth_password=auth_password)

        if not hasattr(self, "get_credentials"):
            raise NotImplementedError(
                "Method 'get_credentials' must be implemented in order to "
                "use 'authenticated_fetch'.")

        update_credentials = self.get_credentials()
        update_credentials(fetch_arguments)

        arguments = {
            "path": fetch_arguments.path
        }

        if method:
            arguments["method"] = method

        if fetch_arguments.body:
            arguments["body"] = fetch_arguments.body
        if fetch_arguments.headers:
            arguments["headers"] = fetch_arguments.headers
        if fetch_arguments.auth_mode:
            arguments["auth_mode"] = fetch_arguments.auth_mode
        if fetch_arguments.auth_password:
            arguments["auth_password"] = fetch_arguments.auth_password
        if fetch_arguments.auth_username:
            arguments["auth_username"] = fetch_arguments.auth_username

        return self.fetch(**arguments)



try:
    from Cookie import SimpleCookie
except ImportError:
    from http.cookies import SimpleCookie

import unittest
from testnado.credentials.cookie_credentials import CookieCredentials
from testnado.credentials.helpers import build_fetch_arguments
from tornado.web import decode_signed_value


class TestCookieCredentials(unittest.TestCase):

    def test_cookie_credentials(self):
        fetch_arguments = build_fetch_arguments("/foobar")
        credentials = CookieCredentials(
            cookie_name="auth", cookie_value="token", cookie_secret="foobar")
        credentials(fetch_arguments)
        self.assertTrue("Cookie" in fetch_arguments.headers)

        cookie = SimpleCookie()
        cookie.load(fetch_arguments.headers["Cookie"])
        self.assertTrue("auth" in cookie)
        cookie_value = cookie["auth"].value
        expected_value = decode_signed_value("foobar", "auth", cookie_value)
        self.assertEqual("token", expected_value.decode("utf-8"))

    def test_cookie_credentials_plaintext(self):
        fetch_arguments = build_fetch_arguments("/foobar")
        credentials = CookieCredentials("auth", "token")
        credentials(fetch_arguments)
        cookie = SimpleCookie()
        cookie.load(fetch_arguments.headers["Cookie"])
        self.assertTrue("auth" in cookie)
        self.assertEqual("token", cookie["auth"].value)



class HeaderCredentials(object):

    def __init__(self, headers):
        self._headers = headers

    def __call__(self, fetch_arguments):
        for header_key, header_value in self._headers.items():
            fetch_arguments.headers.setdefault(header_key, header_value)



from testnado.credentials.helpers import build_fetch_arguments
import unittest


class TestBuildFetchArguments(unittest.TestCase):

    def test_build_fetch_arguments_default(self):
        fetch_arguments = build_fetch_arguments("/foo")
        self.assertEqual("/foo", fetch_arguments.path)
        self.assertEqual({}, fetch_arguments.headers)
        self.assertEqual(None, fetch_arguments.body)
        self.assertEqual(None, fetch_arguments.auth_mode)
        self.assertEqual(None, fetch_arguments.auth_username)
        self.assertEqual(None, fetch_arguments.auth_password)



try:
    from Cookie import SimpleCookie
except ImportError:
    from http.cookies import SimpleCookie

from tornado.web import create_signed_value


class CookieCredentials(object):

    def __init__(self, cookie_name, cookie_value, cookie_secret=None):
        # need to update this so it takes values like domain, expires, etc
        # for people who test more deeply.

        self._cookie = SimpleCookie()
        self._cookie_name = cookie_name
        if cookie_secret:
            cookie_value = create_signed_value(
                cookie_secret, cookie_name, cookie_value).decode("utf-8")
        self._cookie[cookie_name] = cookie_value

    def __call__(self, fetch_arguments):
        cookie_string = self._cookie[self._cookie_name].OutputString(None)
        fetch_arguments.headers.setdefault("Cookie", cookie_string)



# This is used when a mock service can't be constructed, so we intercept
# the fetch. It's not as thorough, since the response object is not
# complete, the network behavior different, etc. but sometimes you
# just can't change internal fetching parameters for dependent libraries.

import contextlib
try:
    import unittest.mock as mock
except ImportError:
    import mock

from tornado import gen
from tornado.httpclient import AsyncHTTPClient, HTTPError
from tornado.httputil import HTTPHeaders


class MockClient(object):

    def __init__(self, ioloop):
        self.ioloop = ioloop
        self.mocked_urls = {}
        self.ioloop.make_current()
        self.client = AsyncHTTPClient()
        self.original_fetch = self.client.fetch

    def mock_url(self, url, method="GET"):
        mock_response = MockResponse(url)
        base_url = url.split("?")[0]
        self.mocked_urls.setdefault(base_url, []).append(
            (method, mock_response))
        return mock_response

    @contextlib.contextmanager
    def patch(self):
        # this is perhaps a bit sketchy -- basing on the idea that
        # AsyncHTTPClient returns the singleton...
        with mock.patch.object(self.client, "fetch", self.fetch):
            yield

    @gen.coroutine
    def fetch(self, url, *args, **kwargs):
        base_url = url.split("?")[0]

        if base_url not in self.mocked_urls:
            response = yield self.original_fetch(url, *args, **kwargs)
            raise gen.Return(response)

        responses = self.mocked_urls[base_url]

        if len(responses) == 0:
            raise MissingMockResponse(
                "URL requested too many times: {}".format(base_url))

        expected_method, response = responses.pop(0)
        request_method = kwargs.get("method", "GET")

        if expected_method != request_method:
            responses.insert(0, (expected_method, response))
            response = MockResponse(url)
            response.body = \
                "Method mismatch ({}) for mocked URL: {} {}".format(
                    request_method, expected_method, base_url)
            response.code = 405

        if response.code > 399 and kwargs.get("raise_error", True):
            raise HTTPError(
                code=response.code, message="Mock error: ({}) {}".format(
                    response.code, response.body), response=response)
        raise gen.Return(response)


class MockResponse(object):

    def __init__(self, url):
        self.code = 200
        self._body = ""
        self._headers = HTTPHeaders()
        self._custom_headers = {}

    @property
    def body(self):
        return self._body

    @body.setter
    def body(self, value):
        self._body = value

    @property
    def headers(self):
        self._headers.setdefault("Content-Length", len(self._body))
        return self._headers


class MissingMockResponse(Exception):
    pass



import collections


_FetchArguments = collections.namedtuple(
    "FetchArguments", [
        "path", "headers", "body", "auth_mode",
        "auth_password", "auth_username"
    ])


def build_fetch_arguments(path, **kwargs):
    kwargs["headers"] = kwargs.get("headers") or {}
    kwargs.setdefault("body", None)
    kwargs.setdefault("auth_mode", None)
    kwargs.setdefault("auth_password", None)
    kwargs.setdefault("auth_username", None)
    return _FetchArguments(path=path, **kwargs)



from testnado.credentials.header_credentials import HeaderCredentials
from testnado.credentials.helpers import build_fetch_arguments
import unittest


class TestHeaderCredentials(unittest.TestCase):

    def test_header_credentials(self):
        fetch_arguments = build_fetch_arguments("/foo")

        credentials = HeaderCredentials({
            "X-Auth-Token": "foobar"
        })

        credentials(fetch_arguments)
        self.assertEqual("foobar", fetch_arguments.headers["X-Auth-Token"])

    def test_header_credentials_existing_headers(self):
        fetch_arguments = build_fetch_arguments(
            "/foo", headers={"X-Auth-Token": "whatever"})
        credentials = HeaderCredentials({
            "X-Auth-Token": "foobar",
        })
        credentials(fetch_arguments)
        self.assertEqual("whatever", fetch_arguments.headers["X-Auth-Token"])



from testnado import AuthenticatedFetchCase
from tornado.testing import AsyncHTTPTestCase

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse


class HandlerTestCase(AuthenticatedFetchCase, AsyncHTTPTestCase):

    def assert_redirected_path_equals(self, expected_path, response):
        if "Location" not in response.headers:
            self.fail("Response does not have a 'Location' header.")
        location = response.headers["Location"]
        path = urlparse.urlparse(location).path
        self.assertEqual(expected_path, path)



import collections
import re

try:
    import urlparse
except ImportError:
    import urllib.parse as urlparse

from tornado.httpserver import HTTPServer
from tornado.web import Application, HTTPError, RequestHandler


try:
    basestring
except NameError:
    basestring = str


class MockService(object):

    def __init__(self, ioloop, port):
        self.ioloop = ioloop
        if isinstance(port, tuple):
            self.socket, self.port = port
        else:
            self.socket = None
            self.port = port
        self.host = "localhost:{0}".format(self.port)
        self.protocol = "http"
        self.base_url = "http://" + self.host
        self.routes = {}
        self._listening = False
        self._service = None

    def url(self, path):
        return urlparse.urljoin(self.base_url, path)

    def listen(self):
        # this is a bit annoying for the end user, but it's better than
        # "magically" starting the service, potentially before the user
        # has finished populating routes (or alternatively never starting)

        if not self.routes.items():
            # making a 'catchall', since without routes Tornado infinitely
            # redirects. this is only really useful for mocking without
            # testing responses (which is sketchy anyway...)
            handler = self.add_method("GET", "/(.*)", _unimplemented)
            handler.add_method("POST", _unimplemented)
            handler.add_method("PUT", _unimplemented)
            handler.add_method("OPTIONS", _unimplemented)
            handler.add_method("INFO", _unimplemented)

        application = Application(self.routes.items())
        if self.socket is not None:
            self._service = HTTPServer(application)
            self._service.add_socket(self.socket)
        else:
            self._service = application.listen(self.port)
        self._listening = True

    def stop(self):
        self._service.stop()

    def add_method(self, method, route, method_handler):
        # this only works with text (not regex) routes, but it's a helper
        # anyway so deal with it. :)
        handler = self.routes.setdefault(route, build_handler(route))
        handler.add_method(method, method_handler)
        return handler

    def assert_requested(self, method, path, headers=None):
        headers = headers or {}
        for handler in self.routes.values():
            if handler.route.match(path):
                try:
                    return handler.assert_requested(method, path, headers)
                except AssertionError:
                    continue
        raise AssertionError("No request matched: {} {}".format(method, path))

    def assert_not_requested(self, method, path, headers=None):
        try:
            self.assert_requested(method=method, path=path, headers=headers)
        except AssertionError:
            pass
        else:
            raise AssertionError(
                "Request was made to: {} {}".format(method, path))


class MockServiceMethods():

    def get(self, *args, **kwargs):
        return self._handle_method("GET", args, kwargs)

    def head(self, *args, **kwargs):
        return self._handle_method("HEAD", args, kwargs)

    def put(self, *args, **kwargs):
        return self._handle_method("PUT", args, kwargs)

    def post(self, *args, **kwargs):
        return self._handle_method("POST", args, kwargs)

    def delete(self, *args, **kwargs):
        return self._handle_method("DELETE", args, kwargs)

    def options(self, *args, **kwargs):
        return self._handle_method("OPTIONS", args, kwargs)

    def info(self, *args, **kwargs):
        return self._handle_method("INFO", args, kwargs)

    def _handle_method(self, method, args, kwargs):
        self.requests.append(self.request)
        if method not in self.method_handlers:
            raise HTTPError(405, "Method '{}' has no handler.".format(method))
        return self.method_handlers[method](self, *args, **kwargs)

    @classmethod
    def add_method(cls, method, handler):
        cls.method_handlers[method.upper()] = handler

    @classmethod
    def assert_requested(cls, method, path, headers):
        for request in cls.requests:
            if not request.method == method:
                continue
            if not request.path == path:
                continue
            if not compare_dicts(
                    headers, request.headers, case_insensitive=True):
                continue

            if request.method == method and request.path == path:
                return request
        raise AssertionError("No request matched: {}".format(method))


def build_handler(handler_route):

    if isinstance(handler_route, basestring):
        handler_route = re.compile(handler_route)

    class Handler(MockServiceMethods, RequestHandler):
        route = handler_route
        requests = []
        method_handlers = {}

    return Handler


_Request = collections.namedtuple("Request", ["method", "path", "headers"])


def _unimplemented(handler, route):
    handler.set_status(501)
    handler.finish("UNIMPLEMENTED RESPONSE")


def compare_dicts(expected, actual, case_insensitive=False):
    if case_insensitive:
        expected = [(k.lower(), v) for k, v in expected.items()]
        actual = [(k.lower(), v) for k, v in actual.items()]
    return set(expected).issubset(actual)



from tornado.testing import bind_unused_port
from testnado.mock_service import MockService


class ServiceCaseHelpers(object):
    # this mixin must be used with an AsyncHTTPTestCase or AsyncTestCase

    @property
    def mock_services(self):
        if not hasattr(self, "_mock_services"):
            self._mock_services = []
        return self._mock_services

    def add_service(self, service=None):
        if not service:
            service = MockService(self.io_loop, bind_unused_port())
        self.mock_services.append(service)
        return service

    def start_services(self):
        for service in self.mock_services:
            service.listen()

    def stop_services(self):
        for service in self.mock_services:
            service.stop()



#!/usr/bin/env python

import sys
from setuptools import setup, find_packages


with open("VERSION") as version_fp:
    VERSION = version_fp.read().strip()


INSTALL_REQUIRES = ["tornado>=4"]

if sys.version_info[0] == 2:
    INSTALL_REQUIRES.append("mock")

setup(
    name="testnado",
    version=VERSION,
    description="Tornado test helpers and Selenium testing with Tornado",
    author='Josh Marshall',
    author_email='catchjosh@gmail.com',
    url="http://github.com/joshmarshall/testnado/",
    license="http://www.apache.org/licenses/LICENSE-2.0",
    packages=find_packages(exclude=["tests", "dist"]),
    install_requires=INSTALL_REQUIRES
)
